var WifiCollector = {};

var AP = require('./AP');
var Client = require('./Client');
const fs = require('fs');
const readline = require('readline');

var aps = new Map();
var clients = new Map();
var ssids = new Map();


  var classicClientColumnNames = new Set();
  classicClientColumnNames.add("Station MAC");
  classicClientColumnNames.add("First time seen");
  classicClientColumnNames.add("Last time seen");
  classicClientColumnNames.add("Power");
  classicClientColumnNames.add("# packets");
  classicClientColumnNames.add("BSSID");
  classicClientColumnNames.add("Probed ESSIDs");

  var newClientColumnNames = new Set();
  newClientColumnNames.add("mac");
  newClientColumnNames.add("first");
  newClientColumnNames.add("last");
  newClientColumnNames.add("pwr");
  newClientColumnNames.add("packets");
  newClientColumnNames.add("ssid");
  newClientColumnNames.add("probed");

  var classicApColumnNames = new Set();
  classicApColumnNames.add("BSSID");
  classicApColumnNames.add("First time seen");
  classicApColumnNames.add("Last time seen");
  classicApColumnNames.add("channel");
  classicApColumnNames.add("Speed");
  classicApColumnNames.add("Privacy");
  classicApColumnNames.add("Cipher");
  classicApColumnNames.add("Authentication");
  classicApColumnNames.add("Power");
  classicApColumnNames.add("# beacons");
  classicApColumnNames.add("# IV");
  classicApColumnNames.add("LAN IP");
  classicApColumnNames.add("ID-length");
  classicApColumnNames.add("ESSID");
  classicApColumnNames.add("Key");

  var newApColumnNames = new Set();
  newApColumnNames.add("mac");
  newApColumnNames.add("first");
  newApColumnNames.add("last");
  newApColumnNames.add("channel");
  newApColumnNames.add("speed");
  newApColumnNames.add("security");
  newApColumnNames.add("cipher");
  newApColumnNames.add("authentication");
  newApColumnNames.add("pwr");
  newApColumnNames.add("beacons");
  newApColumnNames.add("iv");
  newApColumnNames.add("ip");
  newApColumnNames.add("idLength");
  newApColumnNames.add("ssid");
  newApColumnNames.add("hidden");
  newApColumnNames.add("key");

  var clientColumnNames = {
    "Station MAC":"mac",
    "First time seen": "first",
    "Last time seen": "last",
    "Power": "pwr",
    "# packets": "packets",
    "BSSID" : "ap",
    "Probed ESSIDs" : "probed",
  }
  var apColumnNames = {
    "BSSID":"mac",
    "First time seen": "first",
    "Last time seen": "last",
    "channel": "channel",
    "Speed": "speed",
    "Privacy": "security",
    "Cipher": "cipher",
    "Authentication": "authentication",
    "Power": "pwr",
    "# beacons": "beacons",
    "# IV": "iv",
    "LAN IP": "ip",
    "ID-length": "idLength",
    "ESSID": "ssid",
    "Key": "key",
  }

var processFile = function(file, callback) {
  var state = "begin";
  var columns = [];

  var rs = fs.createReadStream(file, { autoClose: true });
  var rl = readline.createInterface({
    input: rs,
  });
  rl.on('line', function(line) {
    switch (state) {
      case "begin":
        if (line.indexOf("BSSID") == 0) {
          columns = line.split(', ');
          state = "aps";
        }
        break;
      case "aps":
        if (line.length<1) {
          state = "between";
          break;
        }
        var entr = line.split(',');
        var ap = new Map();
        for (i=0; i<entr.length; i++) {
          var val = entr[i].replace(/\s+/,"");
          if (val!='' || val!=null) {
            ap.set(apColumnNames[columns[i]], val);
          }
        }
        if (ap.get("ssid").length>1) {
          ssids.set(ap.get("ssid"), ap);
        }
        AP.setAP(aps, ap);
        break;
      case "between":
        if (line.indexOf("Station") == 0) {
          columns = line.split(', ');
          state = "clients";
        }
        break;
      case "clients":
        if (line.length<1) {
          // ende
          endRead();
          break;
        }
        var entr = line.split(',');
        var cli = new Map();
        var p = [];
        for (i=0; i<entr.length; i++) {
          var val = entr[i].replace(/\s+/,'');
          if (val!="" && val!=null && val.length>0) {
            if (i>columns.length-2) { // a probed
              p.push(val);
              ssids.set(val, new Set(["probed"]));
            } else {
              cli.set(clientColumnNames[columns[i]], val);
            }
          }
        }
        var ps = new Set(p);
        cli.set("probed", ps);
        Client.setCli(clients, cli);
        break;
      default:
        console.error("unknown state "+state);
    }
  });
  rs.on('end', function() {
    endRead();
  });
  var endRead = function() {
    rl.close();
    rs.close();
    return callback(null);
  }
};

var importFile = function(file, callback) {
  processFile(file, callback);
}

var exportFile = function(file, callback) {
  var saveClassic = function(file, callback) {
    var classic = "\n";
    classic += "BSSID, First time seen, Last time seen, channel, Speed, Privacy, Cipher, Authentication, Power, # beacons, # IV, LAN IP, ID-length, ESSID, Key";
    classic+="\n";
    for (var v of aps.values()) {
      classic += v.mac+","+v.first+","+v.last+","+v.channel+","+v.speed+","+v.security+","+v.cipher+","+v.authentication+","+v.pwr+","+v.beacons+","+v.iv+","+v.ip+",";
      classic += v.idLength+","+v.ssid+","+v.key+"\n";
    }

    classic+="\n";

    classic += "Station MAC, First time seen, Last time seen, Power, # packets, BSSID, Probed ESSIDs";
    classic+="\n";
    for (var v of clients.values()) {
      classic += v.mac+","+v.first+","+v.last+","+v.pwr+","+v.packets+","+v.ap+",";
      for (var k of v.probed) {
        classic += k+",";
      }
      classic += "\n";
    }
    classic += "\n";


    file = file.replace(/\.csv/,"-classic.csv");
    fs.writeFile(file, classic, function(err) {
      if (err) {
        console.log("writeFile error");
        return callback(err);
      } else {
        saveNew(file, callback);
      }
    });
  }

  var saveNew = function(file, callback) {
    var newT = "";
    for (let v of newApColumnNames.values()) {
      newT += v+",";
    }
    newT += "\n";
    for (let v of newClientColumnNames.values()) {
      newT += v+",";
    }
    newT += "\n";

    file = file.replace(/\.csv/,"-new.csv");
    fs.writeFile(file, newT, function(err) {
      if (err) {
        console.log("writeFile error");
        return callback(err);
      } else {
        return callback();
      }
    });
  }


  saveClassic(file, callback);
}

var print = function() {
  console.log();
  console.log("WifiCollector");
  console.log("SSIDs");
  for (let s of ssids.keys()) {
    console.log(s);
  }
};


WifiCollector.aps = aps;
WifiCollector.clients = clients;
WifiCollector.print = print;
WifiCollector.processFile = processFile;
WifiCollector.importFile = importFile;
WifiCollector.exportFile = exportFile;

module.exports = WifiCollector;
