#! /usr/bin/env node


const term = require('terminal-kit').terminal;
const child_process = require('child_process');
const readline = require('readline');
const fs = require('fs');
const exec = child_process.exec;
const spawn = child_process.spawn;

const WC = require('./WifiCollector');

var standartCallback = function(err, d) {
  if (err) {
    error(err);
  }
}



term.windowTitle("WIFI");
var consoleCursorPos;
term.fullscreen(true);
var termSize = { width: term.width, height: term.height };
term.grabInput({ mouse: 'button' });
var state = "console"; // console | console-tab | log | main


var currCommand = "";
var commands = [ "start", "stop" ];
var commandHistory = new Set();
var commandCursorPos = 0;
var commandIndex = 0;
var currCommandBckp = "";
var commandInfo = "";

var commandAutoComplete = ["prep", "clean", "startMon", "startDump", "startAnalyse", "stopAnalyse", "stopDump", "stopMon", "analyse", "exportWC"];
var currTabMenu = [];
var currTabIndex = 0;

var logEntries = [".",".",".", "logStart"];
var logIndex = 0;




term.on('key', function(name, matches, data) {
  if (name === "CTRL_C") {
    exit();
  } else if (name === "CTRL_L") {
    state = "log";
  } else if (name === "CTRL_M") {
    state = "main";
  } else if(state === "console") {
    if (data.isCharacter) {
      commandInfo = "";
      commandIndex = 0;
      modifyCurrCommand("insert", { char: name });
    } else {
      switch(name) {
        case "ENTER":
          executeCurrCommand();
          break;
        case "BACKSPACE":
          commandIndex = 0;
          commandInfo = "";
          modifyCurrCommand("back-delete");
          break;
        case "DELETE":
          commandIndex = 0;
          commandInfo = "";
          modifyCurrCommand("delete");
          break;
        case "LEFT":
          if (commandCursorPos>0) {
            term.move(-1,0);
            commandCursorPos--;
          }
          commandIndex = 0;
          commandInfo = "";
          break;
        case "RIGHT":
          if (commandCursorPos<currCommand.length) {
            term.move(1,0);
            commandCursorPos++;
          }
          commandIndex = 0;
          commandInfo = "";
          break;
        case "UP":
          commandInfo = "";
          if (commandHistory.size>0) {
            if (commandIndex === 0) {
              currCommandBckp = currCommand;
            }
            commandIndex++;
            var v = Array.from(commandHistory);
            commandIndex = Math.min(v.length, commandIndex);
            setCurrCommand(v[v.length-commandIndex]);
            term.moveTo(commandCursorPos+11,1);
          }
          break;
        case "DOWN":
          commandInfo = "";
          if (commandHistory.size>0) {
            if (commandIndex === 1) {
              setCurrCommand(currCommandBckp);
              commandIndex--;
              term.moveTo(commandCursorPos+11,1);
            } else if (commandIndex>1){
              commandIndex--;
              var v = Array.from(commandHistory);
              commandIndex = Math.max(0, commandIndex);
              setCurrCommand(v[v.length-commandIndex]);
              term.moveTo(commandCursorPos+11,1);
            } else {
            }
          }
          break;
        case "TAB":
          currTabMenu = [];
          currCommandBckp = currCommand;
          commandAutoComplete.map(function(e) {
            if (e.startsWith(currCommand)) {
              currTabMenu.push(e);
            }
          });
          currTabIndex = 0;
          if (currTabMenu.length>0) {
            state = "console-tab";
            setCurrCommand(currTabMenu[currTabIndex]);
          } else {

          }
          //log(currTabMenu);
          break;
        default:
          log("pressed "+name);
      }
    }
    //console.log("key event ", name, data);
  } else if (state === "console-tab") { // TODO
    switch(name) {
      case "UP": // go up in list
        currTabIndex = Math.max(0,currTabIndex-1);
        setCurrCommand(currTabMenu[currTabIndex]);
        break;
      case "DOWN": // go down in list
        currTabIndex = Math.min(currTabMenu.length-1,currTabIndex+1);
        setCurrCommand(currTabMenu[currTabIndex]);
        break;
      case "ENTER": // select entry and run
        state = "console";
        executeCurrCommand();
        break;
      case "TAB":
        currTabIndex = currTabIndex+1;
        if (currTabIndex>currTabMenu.length-1) {
          currTabIndex = 0;
        }
        setCurrCommand(currTabMenu[currTabIndex]);
        break;
      default:  // if special caracter (like ESC) cancel
                // otherwise select entry and write additional stuff
        if (!data.isCharacter) {
          setCurrCommand(currCommandBckp);
          state = "console";
        } else {
          setCurrCommand(currCommand+name);
          state = "console";
        }
    }
  } else if (state === "log") {
    switch(name) {
      case "UP":
        logIndex++;
        break;
      case "DOWN":
        logIndex--;
        break;
      case "LEFT":
        logIndex+=10;
        break;
      case "RIGHT":
        logIndex-=10;
        break;
      default:
        state = "console";
        if (data.isCharacter) {
          setCurrCommand(currCommand+name);
        } else {
          setCurrCommand(currCommand);
        }
    }
  }
  drawLayout();
});
term.on('resize', function(w, h) {
  termSize.width = w;
  termSize.height = h;
  drawLayout();
});
term.on('terminal', function(name, data) {
  if (name ==="SCREEN_RESIZE") {
  } else {
    console.log("terminal event: "+name,data);
  }
});
term.on('mouse', function(name, data) {
  switch(name) {
    case "MOUSE_LEFT_BUTTON_PRESSED":
      if (data.y === 1) {
        term.hideCursor(false);
        state = "console";
      } else if (data.y > 2 && data.y < 7) {
        term.hideCursor();
        state = "log";
      } else if (data.y > 7) {
        state = "main";
      }
      break;
    default:
  }
  //console.log("mouse event: ",name,data);
});

var setCurrCommand = function(cmd) {
  currCommand = cmd;
  commandCursorPos = cmd.length;
}

var drawLayout = function() {
  term.hideCursor();
  term.bgColorGrayscale.clear(0);

  // draw log
  term.moveTo.blue(1,2, "log")
  for (let i=4; i<=termSize.width; i++) {
    term.moveTo.brightBlack(i,2, "-");
  }
  drawLog();

  // draw main
  term.moveTo.brightYellow(1,7, "main");
  for (let i=5; i<=termSize.width; i++) {
    term.moveTo.brightBlack(i,7, "-");
  }

  // draw console
  term.moveTo.green(1,1, "console >");
  drawCurrCommand();

}

var drawCurrCommand = function() {
  if (currCommand.length === 0){
    term.moveTo.eraseLineAfter.green(11,1, commandInfo); // TODO: commandCursorPos is not correct when text out of screen and cursor not at end of line
  } else if (currCommand.length > (termSize.width-11)) {
    term.moveTo.eraseLineAfter.brightGreen(11,1, currCommand.substring(currCommand.length-termSize.width+11));
  } else {
    term.moveTo.eraseLineAfter.brightGreen(11,1, currCommand);
  }
  term.moveTo(11+commandCursorPos,1);
  if (state === "console") {
    term.hideCursor(false);
  }
}

var drawLog = function() {
  term.moveTo(5,2, ""+(Math.max(logEntries.length-7,0)));
  logIndex = Math.max(Math.min(logEntries.length-7, logIndex),0);
  term.moveTo(15,2, ""+logIndex);
  for (let i=0; i<4; i++) {
    term.moveTo(1,3+i, " "+logEntries[logEntries.length-4+i-logIndex]);
  }
}

var modifyCurrCommand = function(op, d) {
  switch (op) {
    case "insert":
      var a = currCommand.substring(0, commandCursorPos+1);
      if (d.pos>currCommand.length-1) {
        var b = "";
      } else {
        var b = currCommand.substring(commandCursorPos);
      }
      setCurrCommand(a+d.char+b);
      break;
    case "back-delete":
      var a = currCommand.substring(0,commandCursorPos-1);
      var b = currCommand.substring(commandCursorPos);
      setCurrCommand(a+b);
      break;
    case "delete":
      var a = currCommand.substring(0,commandCursorPos);
      var b = currCommand.substring(commandCursorPos+1);
      setCurrCommand(a+b);
      break;
  }
}

var exit = function() {
  clearInterval(drawInt);
  term.bgColorGrayscale.clear.moveTo(0,3,3, "closing ...");
  setTimeout(function() {
    term.fullscreen(false);
    term.grabInput(false);
    term.processExit(0);
  },0);
  //},1000);
}

var executeCurrCommand = function() {
  var command = "";
  if (currCommand.length>0) {
    commandIndex = 0;
    commandHistory.delete(currCommand);
    commandHistory.add(currCommand);
    command = currCommand;
    currCommand = "";
    commandCursorPos = 0;
  }
  command = command.split(" ");
  switch(command[0]) {
    case "log":
      log(command.join(" "));
      break;
    case "error":
      error(command.join(" "))
      break;
    case "warn":
      warn(command.join(" "));
      break;
    default:
      if (AIR[command[0]]) {
        AIR[command[0]](command);
      } else {
        commandInfo = "command not found";
        var v = Array.from(commandHistory);
        commandHistory.delete(v[v.length-1]);
      }
  }
}

var log = function(what) {
  if (logEntries.length>1000) { logEntries.shift(); }
  logEntries.push(term.brightBlue.str(what));
}

var warn = function(what) {
  logIndex = 0;
  logEntries.push(term.yellow.str(what));
}

var error = function(what) {
  logIndex = 0;
  logEntries.push(term.red.str(what));
}























// aircrack part

var AIR = {};
var analyseInterval;

AIR.startMon = function(c, callback) {
  if (callback == null) { callback = standartCallback }
  exec("airmon-ng", function(err, stdout, stderr) {
    if (stdout.indexOf("mon0")>-1) {
      warn("airmon already started");
    } else {
      var interf = "";
      var ff = stdout;
      ff = ff.split("\n");
      for (let i=0; i<ff.length; i++) {
        if (ff[i].length>0 && ff[i].indexOf("Interface")<0) {
          var interf = ff[i].replace(/\s.*/, "");
        }
      }
      exec("airmon-ng start "+interf, function(err, stdout, stderr) {
        if (!err) {
          log("airmon started");
          return callback(null);
        } else {
          error("airmon failed to run");
          return callback(err);
        }
      });

    }
  });
}
AIR.startDump = function(c, callback) { // TODO: recognise if prep is required
  if (callback == null) { callback = standartCallback }
  var outputFile = "./wI-tmp/";
  if (c[1] != null) {
    outputFile += c[1];
  } else {
    outputFile += "outputFile";
  }
  exec("airodump-ng -w "+outputFile+" --output-format csv mon0 &", function(err, stdout, stderr) {
    //log(stderr);
    log("airodump started");
  });

  return callback(null);
}

AIR.stopDump = function(c, callback) {
  if (callback == null) { callback = standartCallback }
  var ps = spawn("ps", ["aux"]);
  var psRl = readline.createInterface({
    input: ps.stdout
  });
  psRl.on('line', function(data) {
    if (data.indexOf("airodump-ng")>0 ) {
      //console.log(data.substring(9,14));
      var kill = spawn("kill", ["-SIGKILL", data.substring(9,14)]);
    }
  });
  ps.stdout.on('end', function() {
    log("killed airodump");
    psRl.close();
  });
}
AIR.stopMon = function(c, callback) {
  if (callback == null) { callback = standartCallback }
  exec("airmon-ng stop mon0", function(err, stdout, stderr) {
    if (!err) {
      log("airmon stopped");
      exec("airmon-ng stop mon1");
    } else {
      error("problem stopping airmon-ng");
    }
  });
}
AIR.prep = function(c, callback) {
  if (callback == null) { callback = standartCallback }
  AIR.clean(null, function(err) {
    fs.mkdir("./wI-tmp/", function(err) {
      AIR.getOutputFile(function(err, file) {
        if (err) {
          // no outputFile yet
        } else {
          // outputFile to delete (or move)
          fs.unlink(file, function(err) {
            if (err) {
              error("Error deleting outputFile\n"+err);
              return callback(new Error("Error deleting outputFile"));
            }
          });
        }
      });
    });
  });
}
AIR.getOutputFile = function(c, callback) {
  if (callback == null) { callback = standartCallback; }
  var filename = "";
  if (c[1] != null) {
    filename += c[1];
  } else {
    filename += "outputFile";
  }
  var f = null;
  fs.readdir("./wI-tmp/", function(err, files) {
    if (err) {
      error("Error getting files in folder");
      return callback(new Error("Error getting files in folder"));
    } else {
      for (let i=files.length-1; i>=0; i--) {
        var file = files[i];
        if (file.indexOf(filename)>-1) {
          f = "./wI-tmp/"+file;
          i=files.length;
          return callback(null, f);
        }
      }
      return callback(new Error("Error: no outputFile called "+filename+" yet"));
    }
  });
}
AIR.clean = function(c, callback) {
  if (callback == null) { callback = standartCallback }
  exec("rm -rf ./wI-tmp/", function(err, stdout, stderr) {
    if (err) {
      error(err);
      callback(err);
    } else {
      warn("cleaned current workingdir");
      return callback(null);
    }
  });
}
AIR.analyse = function(c, callback) {
  if (callback == null) { callback = standartCallback; }
  var filename = "";
  if (c[1] != null){
    filename = "outputFile";
  } else {
    filename = c[1];
  }
  AIR.getOutputFile([null,filename], function(err, file) {
    if (c[1]) {
      file = "./wI-tmp/"+c[1];
      WC.processFile(file, callback);
    } else if (err) {
      error("no outputFile, start dump with startDump or specify file as first argument 'analyse file'");
      return callback(err);
    } else {
      log("processing file");
      WC.processFile(file, callback);
    }
  });
}
AIR.startAnalyse = function(c, callback) {
  if (callback == null) { callback = standartCallback; }
  var t = 10*1000
  if (c[1]!=null) {
    t = parseInt(c[1]);
  }
  t = Math.max(t, 5*1000);
  analyseInterval = setInterval(AIR.analyse, t, []);
  log("started analysing interval");
  return callback(null);
}
AIR.stopAnalyse = function(c, callback) {
  if (callback == null) { callback = standartCallback; }
  clearInterval(analyseInterval);
  log("stopped analysing");
  return callback(null);
}
AIR.exportWC = function (c, callback) {
  if (callback == null) { callback = standartCallback; }
  WC.exportFile("./export.csv", function(err) {
    if (err) {
      error(err);
    } else {
      log("exported to ./export.csv");
      return callback(null);
    }
  });
}
AIR.print = function(c, callback) {
  if (callback == null) { callback = standartCallback; }
  for (let [k,v] of WC.aps.entries()) {
    log(k);
  }
}








// start
drawLayout();
var drawInt = setInterval(drawLayout, 1000);
